@@ -19,18 +19,28 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-import time
 import copy
 from firewall.core.base import SHORTCUTS, DEFAULT_ZONE_TARGET, SOURCE_IPSET_TYPES
 from firewall.core.fw_transaction import FirewallTransaction
 from firewall.core.io.policy import Policy
 from firewall.core.logger import log
-from firewall.core.rich import Rich_Service, Rich_Port, Rich_Protocol, Rich_SourcePort, Rich_ForwardPort, \
-                               Rich_IcmpBlock, Rich_IcmpType, Rich_Masquerade, Rich_Mark, Rich_Tcp_Mss_Clamp
+from firewall.core.rich import (
+    Rich_ForwardPort,
+    Rich_IcmpBlock,
+    Rich_IcmpType,
+    Rich_Mark,
+    Rich_Masquerade,
+    Rich_Port,
+    Rich_Protocol,
+    Rich_Rule,
+    Rich_Service,
+    Rich_SourcePort,
+    Rich_Tcp_Mss_Clamp
+)
+from firewall.core.fw_nm import nm_get_bus_name
 from firewall.functions import checkIPnMask, checkIP6nMask, check_mac
 from firewall import errors
 from firewall.errors import FirewallError
-from firewall.fw_types import LastUpdatedOrderedDict
 
 class FirewallZone(object):
     ZONE_POLICY_PRIORITY = 0
@@ -48,7 +58,9 @@ class FirewallZone(object):
         self._zone_policies.clear()
 
     def new_transaction(self):
-        return FirewallTransaction(self._fw)
+        t = FirewallTransaction(self._fw)
+        t.add_pre(self._fw.full_check_config)
+        return t
 
     def policy_name_from_zones(self, fromZone, toZone):
         return "zone_{fromZone}_{toZone}".format(fromZone=fromZone, toZone=toZone)
@@ -58,17 +70,19 @@ class FirewallZone(object):
     def get_zones(self):
         return sorted(self._zones.keys())
 
-    def get_active_zones(self):
+    def get_active_zones(self, append_default=True):
         active_zones = []
         for zone in self.get_zones():
             if self.list_interfaces(zone) or self.list_sources(zone):
                 active_zones.append(zone)
+        if append_default and self._fw._default_zone not in active_zones:
+            active_zones.append(self._fw._default_zone)
         return active_zones
 
     def get_zone_of_interface(self, interface):
         interface_id = self.__interface_id(interface)
         for zone in self._zones:
-            if interface_id in self._zones[zone].settings["interfaces"]:
+            if interface_id in self._zones[zone].interfaces:
                 # an interface can only be part of one zone
                 return zone
         return None
@@ -76,7 +90,7 @@ class FirewallZone(object):
     def get_zone_of_source(self, source):
         source_id = self.__source_id(source)
         for zone in self._zones:
-            if source_id in self._zones[zone].settings["sources"]:
+            if source_id in self._zones[zone].sources:
                 # a source_id can only be part of one zone
                 return zone
         return None
@@ -100,10 +114,11 @@ class FirewallZone(object):
         for setting in ["services", "ports",
                         "masquerade", "forward_ports",
                         "source_ports",
-                        "icmp_blocks", "rules",
-                        "protocols"]:
+                        "icmp_blocks", "icmp_block_inversion",
+                        "rules_str", "protocols"]:
             if fromZone == z_obj.name and toZone == "HOST" and \
-               setting in ["services", "ports", "source_ports", "icmp_blocks", "protocols"]:
+               setting in ["services", "ports", "source_ports", "icmp_blocks",
+                           "icmp_block_inversion", "protocols"]:
                 # zone --> HOST
                 setattr(p_obj, setting, copy.deepcopy(getattr(z_obj, setting)))
             elif fromZone == "ANY" and toZone == z_obj.name and setting in ["masquerade"]:
@@ -113,26 +128,26 @@ class FirewallZone(object):
                  setting in ["forward_ports"]:
                 # zone --> any zone
                 setattr(p_obj, setting, copy.deepcopy(getattr(z_obj, setting)))
-            elif setting in ["rules"]:
+            elif setting in ["rules_str"]:
+                p_obj.rules_str = []
                 p_obj.rules = []
-                for rule in z_obj.rules:
+                for rule_str in z_obj.rules_str:
                     current_policy = self.policy_name_from_zones(fromZone, toZone)
 
+                    rule = Rich_Rule(rule_str=rule_str)
                     if current_policy in self._rich_rule_to_policies(z_obj.name, rule):
-                        p_obj.rules.append(copy.deepcopy(rule))
+                        p_obj.rules_str.append(rule_str)
+                        p_obj.rules.append(rule)
 
         return p_obj
 
     def add_zone(self, obj):
-        obj.settings = { x : LastUpdatedOrderedDict()
-                         for x in ["interfaces", "sources",
-                                   "icmp_block_inversion",
-                                   "forward"] }
         self._zones[obj.name] = obj
         self._zone_policies[obj.name] = []
 
         # Create policy objects, will need many:
         #   - (zone --> HOST) - ports, service, etc
+        #   - (HOST --> zone) - tracking only, no zone features
         #   - (any zone --> zone) - masquerade
         #   - (zone --> any zone)
         #       - also includes forward-ports because it works on (nat,
@@ -140,31 +155,16 @@ class FirewallZone(object):
         #       host or dnat to a different host.
         #       - also includes rich rule "mark" action for the same reason
         #
-        for fromZone,toZone in [(obj.name, "HOST"),
+        for fromZone,toZone in [(obj.name, "HOST"), ("HOST", obj.name),
                                 ("ANY", obj.name), (obj.name, "ANY")]:
             p_obj = self.policy_obj_from_zone_obj(obj, fromZone, toZone)
             self._fw.policy.add_policy(p_obj)
             self._zone_policies[obj.name].append(p_obj.name)
 
-        self.copy_permanent_to_runtime(obj.name)
-
-    def copy_permanent_to_runtime(self, zone):
-        obj = self._zones[zone]
-
-        for arg in obj.interfaces:
-            self.add_interface(zone, arg, allow_apply=False)
-        for arg in obj.sources:
-            self.add_source(zone, arg, allow_apply=False)
-        if obj.forward:
-            self.add_forward(zone)
-        if obj.icmp_block_inversion:
-            self.add_icmp_block_inversion(zone)
-
     def remove_zone(self, zone):
         obj = self._zones[zone]
         if obj.applied:
             self.unapply_zone_settings(zone)
-        obj.settings.clear()
         del self._zones[zone]
         del self._zone_policies[zone]
 
@@ -241,31 +241,20 @@ class FirewallZone(object):
                 if use_transaction is None:
                     transaction.execute(True)
 
-    # settings
-
-    # generate settings record with sender, timeout
-    def __gen_settings(self, timeout, sender):
-        ret = {
-            "date": time.time(),
-            "sender": sender,
-            "timeout": timeout,
-        }
-        return ret
-
-    def get_settings(self, zone):
-        return self.get_zone(zone).settings
-
     def _zone_settings(self, enable, zone, transaction):
-        settings = self.get_settings(zone)
-        for key in settings:
-            for args in settings[key]:
+        for key in ["interfaces", "sources", "forward", "icmp_block_inversion"]:
+            args_list = getattr(self.get_zone(zone), key)
+            if isinstance(args_list, bool):
+                args_list = [args_list]
+            for args in args_list:
                 if key == "interfaces":
                     self._interface(enable, zone, args, transaction)
                 elif key == "sources":
-                    self._source(enable, zone, args[0], args[1], transaction)
+                    ipv = self.check_source(args)
+                    self._source(enable, zone, ipv, args, transaction)
                 elif key == "icmp_block_inversion":
                     continue
-                elif key == "forward":
+                elif key in ["forward", "ingress_priority", "egress_priority"]:
                     # no need to call this when applying the zone as the rules
                     # will be generated when adding the interfaces/sources
                     pass
@@ -356,7 +345,6 @@ class FirewallZone(object):
 
     def set_config_with_settings_dict(self, zone, settings, sender):
         # stupid wrappers to convert rich rule string to rich rule object
-        from firewall.core.rich import Rich_Rule
         def add_rule_wrapper(zone, rule_str, timeout=0, sender=None):
             self.add_rule(zone, Rich_Rule(rule_str=rule_str), timeout=0, sender=sender)
         def remove_rule_wrapper(zone, rule_str):
@@ -377,6 +365,13 @@ class FirewallZone(object):
             "forward": (self.add_forward, self.remove_forward),
         }
 
+        # do a full config check on a temporary object before trying to make
+        # the runtime changes
+        old_obj = self.get_zone(zone)
+        check_obj = copy.copy(old_obj)
+        check_obj.import_config_dict(settings, self._fw.get_all_io_objects_dict())
+        self._fw.full_check_config({"zones": [check_obj]})
+
         old_settings = self.get_config_with_settings_dict(zone)
         (add_settings, remove_settings) = self._fw.get_added_and_removed_settings(old_settings, settings)
 
@@ -413,18 +408,6 @@ class FirewallZone(object):
     def check_interface(self, interface):
         self._fw.check_interface(interface)
 
-    def interface_get_sender(self, zone, interface):
-        _zone = self._fw.check_zone(zone)
-        _obj = self._zones[_zone]
-        interface_id = self.__interface_id(interface)
-
-        if interface_id in _obj.settings["interfaces"]:
-            settings = _obj.settings["interfaces"][interface_id]
-            if "sender" in settings and settings["sender"] is not None:
-                return settings["sender"]
-
-        return None
-
     def __interface_id(self, interface):
         self.check_interface(interface)
         return interface
@@ -437,7 +420,7 @@ class FirewallZone(object):
 
         interface_id = self.__interface_id(interface)
 
-        if interface_id in _obj.settings["interfaces"]:
+        if interface_id in _obj.interfaces:
             raise FirewallError(errors.ZONE_ALREADY_SET,
                                 "'%s' already bound to '%s'" % (interface,
                                                                 zone))
@@ -473,11 +456,11 @@ class FirewallZone(object):
         return _zone
 
     def __register_interface(self, _obj, interface_id, zone, sender):
-        _obj.settings["interfaces"][interface_id] = \
-            self.__gen_settings(0, sender)
-        # add information whether we add to default or specific zone
-        _obj.settings["interfaces"][interface_id]["__default__"] = \
-            (not zone or zone == "")
+        _obj.interfaces.append(interface_id)
+        if not zone or zone == "":
+            self._fw._default_zone_interfaces.append(interface_id)
+        if sender == nm_get_bus_name():
+            self._fw._nm_assigned_interfaces.append(interface_id)
 
     def change_zone_of_interface(self, zone, interface, sender=None):
         self._fw.check_panic()
@@ -494,22 +477,6 @@ class FirewallZone(object):
 
         return _zone
 
-    def change_default_zone(self, old_zone, new_zone, use_transaction=None):
-        self._fw.check_panic()
-
-        if use_transaction is None:
-            transaction = self.new_transaction()
-        else:
-            transaction = use_transaction
-
-        self.apply_zone_settings(new_zone, transaction)
-        self._interface(True, new_zone, "+", transaction, append=True)
-        if old_zone is not None and old_zone != "":
-            self._interface(False, old_zone, "+", transaction, append=True)
-
-        if use_transaction is None:
-            transaction.execute(True)
-
     def remove_interface(self, zone, interface,
                          use_transaction=None):
         self._fw.check_panic()
@@ -539,14 +506,18 @@ class FirewallZone(object):
         return _zone
 
     def __unregister_interface(self, _obj, interface_id):
-        if interface_id in _obj.settings["interfaces"]:
-            del _obj.settings["interfaces"][interface_id]
+        if interface_id in _obj.interfaces:
+            _obj.interfaces.remove(interface_id)
+        if interface_id in self._fw._default_zone_interfaces:
+            self._fw._default_zone_interfaces.remove(interface_id)
+        if interface_id in self._fw._nm_assigned_interfaces:
+            self._fw._nm_assigned_interfaces.remove(interface_id)
 
     def query_interface(self, zone, interface):
-        return self.__interface_id(interface) in self.get_settings(zone)["interfaces"]
+        return self.__interface_id(interface) in self.get_zone(zone).interfaces
 
     def list_interfaces(self, zone):
-        return self.get_settings(zone)["interfaces"].keys()
+        return self.get_zone(zone).interfaces
 
     # SOURCES
 
@@ -566,8 +537,8 @@ class FirewallZone(object):
             raise FirewallError(errors.INVALID_ADDR, source)
 
     def __source_id(self, source, applied=False):
-        ipv = self.check_source(source, applied=applied)
-        return (ipv, source)
+        self.check_source(source, applied=applied)
+        return source
 
     def add_source(self, zone, source, sender=None, use_transaction=None,
                    allow_apply=True):
@@ -578,9 +549,10 @@ class FirewallZone(object):
         if check_mac(source):
             source = source.upper()
 
+        ipv = self.check_source(source, applied=allow_apply)
         source_id = self.__source_id(source, applied=allow_apply)
 
-        if source_id in _obj.settings["sources"]:
+        if source_id in _obj.sources:
             raise FirewallError(errors.ZONE_ALREADY_SET,
                             "'%s' already bound to '%s'" % (source, _zone))
         if self.get_zone_of_source(source) is not None:
@@ -598,7 +570,7 @@ class FirewallZone(object):
             transaction.add_fail(self.set_zone_applied, _zone, False)
 
         if allow_apply:
-            self._source(True, _zone, source_id[0], source_id[1], transaction)
+            self._source(True, _zone, ipv, source_id, transaction)
 
         self.__register_source(_obj, source_id, zone, sender)
         transaction.add_fail(self.__unregister_source, _obj, source_id)
@@ -609,10 +581,7 @@ class FirewallZone(object):
         return _zone
 
     def __register_source(self, _obj, source_id, zone, sender):
-        _obj.settings["sources"][source_id] = \
-            self.__gen_settings(0, sender)
-        # add information whether we add to default or specific zone
-        _obj.settings["sources"][source_id]["__default__"] = (not zone or zone == "")
+        _obj.sources.append(source_id)
 
     def change_zone_of_source(self, zone, source, sender=None):
         self._fw.check_panic()
@@ -653,9 +622,10 @@ class FirewallZone(object):
             transaction = use_transaction
 
         _obj = self._zones[_zone]
+        ipv = self.check_source(source)
         source_id = self.__source_id(source)
         transaction.add_post(self.__unregister_source, _obj, source_id)
-        self._source(False, _zone, source_id[0], source_id[1], transaction)
+        self._source(False, _zone, ipv, source_id, transaction)
 
         if use_transaction is None:
             transaction.execute(True)
@@ -663,58 +633,158 @@ class FirewallZone(object):
         return _zone
 
     def __unregister_source(self, _obj, source_id):
-        if source_id in _obj.settings["sources"]:
-            del _obj.settings["sources"][source_id]
+        if source_id in _obj.sources:
+            _obj.sources.remove(source_id)
 
     def query_source(self, zone, source):
         if check_mac(source):
             source = source.upper()
-        return self.__source_id(source) in self.get_settings(zone)["sources"]
+        return self.__source_id(source) in self.get_zone(zone).sources
 
     def list_sources(self, zone):
-        return [ k[1] for k in self.get_settings(zone)["sources"].keys() ]
+        return self.get_zone(zone).sources
 
-    def _interface(self, enable, zone, interface, transaction, append=False):
+    def _interface_or_source_update_zone_termination(self, enable, zone, interface, source, transaction):
+        self._interface_or_source_update_policies_derived_from_zone(enable, zone, interface, source, transaction, last=True)
+
+    def _interface_or_source_update_policies_derived_from_zone(self, enable, zone, interface, source, transaction, last=False):
+        # host <--> zone
+        policy = self.policy_name_from_zones(zone, "HOST")
+        self._fw.policy._ingress_egress_pair(enable, policy, zone, "HOST",
+                                             interface, source, "", "",
+                                             transaction, last=last)
+        policy = self.policy_name_from_zones("HOST", zone)
+        self._fw.policy._ingress_egress_pair(enable, policy, "HOST", zone,
+                                             "", "", interface, source,
+                                             transaction, last=last)
+        # zone <--> zone
+        _zoneB_list = set(self.get_active_zones() + [zone]) if interface != "+" else [self._fw._default_zone]
+        for _zoneB in _zoneB_list:
+            if _zoneB != zone and not self.get_zone(_zoneB).applied:
+                continue
+
+            _zoneB_interfaces = list(self.list_interfaces(_zoneB))
+            _zoneB_sources = list(self.list_sources(_zoneB))
+            if zone == _zoneB and interface and interface not in _zoneB_interfaces:
+                # include the interface that's currently being activated
+                _zoneB_interfaces.append(interface)
+            if zone == _zoneB and source and source not in _zoneB_sources:
+                _zoneB_sources.append(source)
+            # Need to consider dispatch to the default zone as well.
+            if interface == "+":
+                _zoneB_interfaces = ["+"]
+                _zoneB_sources = []
+            elif _zoneB == self._fw._default_zone and "+" not in _zoneB_interfaces:
+                _zoneB_interfaces.append("+")
+
+            for _zoneB_interface in _zoneB_interfaces:
+                policy = self.policy_name_from_zones(zone, "ANY")
+                self._fw.policy._ingress_egress_pair(enable, policy, zone, _zoneB,
+                                                     interface, source, _zoneB_interface, "",
+                                                     transaction, last=last)
+                policy = self.policy_name_from_zones("ANY", zone)
+                self._fw.policy._ingress_egress_pair(enable, policy, _zoneB, zone,
+                                                     _zoneB_interface, "", interface, source,
+                                                     transaction, last=last)
+                if interface != _zoneB_interface:
+                    policy = self.policy_name_from_zones(_zoneB, "ANY")
+                    self._fw.policy._ingress_egress_pair(enable, policy, _zoneB, zone,
+                                                         _zoneB_interface, "", interface, source,
+                                                         transaction, last=last)
+                    policy = self.policy_name_from_zones("ANY", _zoneB)
+                    self._fw.policy._ingress_egress_pair(enable, policy, zone, _zoneB,
+                                                         interface, source, _zoneB_interface, "",
+                                                         transaction, last=last)
+            for _zoneB_source in _zoneB_sources:
+                # must be same IPv4/IPv6 family!
+                if source and self.check_source(source) != self.check_source(_zoneB_source):
+                    continue
+                policy = self.policy_name_from_zones(zone, "ANY")
+                self._fw.policy._ingress_egress_pair(enable, policy, zone, _zoneB,
+                                                     interface, source, "", _zoneB_source,
+                                                     transaction, last=last)
+                policy = self.policy_name_from_zones("ANY", zone)
+                self._fw.policy._ingress_egress_pair(enable, policy, _zoneB, zone,
+                                                     "", _zoneB_source, interface, source,
+                                                     transaction, last=last)
+                if source != _zoneB_source:
+                    policy = self.policy_name_from_zones(_zoneB, "ANY")
+                    self._fw.policy._ingress_egress_pair(enable, policy, _zoneB, zone,
+                                                         "", _zoneB_source, interface, source,
+                                                         transaction, last=last)
+                    policy = self.policy_name_from_zones("ANY", _zoneB)
+                    self._fw.policy._ingress_egress_pair(enable, policy, zone, _zoneB,
+                                                         interface, source, "", _zoneB_source,
+                                                         transaction, last=last)
+
+    def _interface_or_source_update_policies(self, enable, zone, interface, source, transaction):
+        # update policy dispatch for any policy using this zone as an
+        # ingress-zone or egress-zone
+        for policy in self._fw.policy.get_policies_not_derived_from_zone():
+            if enable and not self._fw.policy.get_policy(policy).applied:
+                transaction.add_post(self._fw.policy.try_apply_policy_settings, policy)
+            elif self._fw.policy.get_policy(policy).applied:
+                if not enable:
+                    transaction.add_post(self._fw.policy.try_unapply_policy_settings, policy)
+
+                ingress_zone_list = self._fw.policy.list_ingress_zones(policy)
+                egress_zone_list = self._fw.policy.list_egress_zones(policy)
+
+                if "ANY" in ingress_zone_list and "ANY" in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif "ANY" in ingress_zone_list and "HOST" in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                elif "HOST" in ingress_zone_list and "ANY" in egress_zone_list:
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif zone in ingress_zone_list and "ANY" in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif zone in ingress_zone_list and "HOST" in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                elif "ANY" in ingress_zone_list and zone in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif "HOST" in ingress_zone_list and zone in egress_zone_list:
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif zone in ingress_zone_list and zone in egress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+                elif zone in ingress_zone_list:
+                    self._fw.policy._ingress_zone(enable, policy, zone, transaction,
+                                                  ingressInterface=interface, ingressSource=source)
+                elif zone in egress_zone_list:
+                    self._fw.policy._egress_zone(enable, policy, zone, transaction,
+                                                 egressInterface=interface, egressSource=source)
+
+    def _interface(self, enable, zone, interface, transaction):
         for backend in self._fw.enabled_backends():
             if not backend.policies_supported:
                 continue
-            for policy in self._zone_policies[zone]:
-                for (table, chain) in self._fw.policy._get_table_chains_for_zone_dispatch(policy):
-                    rules = backend.build_zone_source_interface_rules(enable,
-                                    zone, policy, interface, table, chain, append)
-                    transaction.add_rules(backend, rules)
+            if interface == "+":
+                continue
 
             # intra zone forward
             policy = self.policy_name_from_zones(zone, "ANY")
-            # Skip adding wildcard/catch-all interface (for default
-            # zone). Otherwise it would allow forwarding from interface
-            # in default zone -> interface not in default zone (but in
-            # a different zone).
-            if self.get_settings(zone)["forward"] and interface not in ["+", "*"]:
+            if self.get_zone(zone).forward:
                 rules = backend.build_zone_forward_rules(enable, zone, policy, "filter", interface=interface)
                 transaction.add_rules(backend, rules)
 
-        # update policy dispatch for any policy using this zone in ingress
-        # or egress
-        for policy in self._fw.policy.get_policies_not_derived_from_zone():
-            if zone not in self._fw.policy.list_ingress_zones(policy) and \
-               zone not in self._fw.policy.list_egress_zones(policy):
-                continue
-            if policy in self._fw.policy.get_active_policies_not_derived_from_zone() and self._fw.policy.get_policy(policy).applied:
-                # first remove the old set of interfaces using the current zone
-                # settings.
-                if not enable and len(self.list_interfaces(zone)) == 1:
-                    self._fw.policy.unapply_policy_settings(policy, use_transaction=transaction)
-                else:
-                    self._fw.policy._ingress_egress_zones(False, policy, transaction)
-                    # after the transaction ends and therefore the interface
-                    # has been added to the zone's settings, update the
-                    # dependent policies
-                    transaction.add_post(lambda p: (p in self._fw.policy.get_active_policies_not_derived_from_zone()) and \
-                                                   self._fw.policy._ingress_egress_zones_transaction(True, p), policy)
-            elif enable:
-                transaction.add_post(lambda p: (p in self._fw.policy.get_active_policies_not_derived_from_zone()) and \
-                                               self._fw.policy.apply_policy_settings(p), policy)
+        self._interface_or_source_update_zone_termination(enable, zone, interface, "", transaction)
+        self._interface_or_source_update_policies_derived_from_zone(enable, zone, interface, "", transaction)
+        self._interface_or_source_update_policies(enable, zone, interface, "", transaction)
 
     # IPSETS
 
@@ -746,39 +816,16 @@ class FirewallZone(object):
         for backend in [self._fw.get_backend_by_ipv(ipv)] if ipv else self._fw.enabled_backends():
             if not backend.policies_supported:
                 continue
-            for policy in self._zone_policies[zone]:
-                for (table, chain) in self._fw.policy._get_table_chains_for_zone_dispatch(policy):
-                    rules = backend.build_zone_source_address_rules(enable, zone,
-                                            policy, source, table, chain)
-                    transaction.add_rules(backend, rules)
 
             # intra zone forward
             policy = self.policy_name_from_zones(zone, "ANY")
-            if self.get_settings(zone)["forward"]:
+            if self.get_zone(zone).forward:
                 rules = backend.build_zone_forward_rules(enable, zone, policy, "filter", source=source)
                 transaction.add_rules(backend, rules)
 
-        # update policy dispatch for any policy using this zone in ingress
-        # or egress
-        for policy in self._fw.policy.get_policies_not_derived_from_zone():
-            if zone not in self._fw.policy.list_ingress_zones(policy) and \
-               zone not in self._fw.policy.list_egress_zones(policy):
-                continue
-            if policy in self._fw.policy.get_active_policies_not_derived_from_zone() and self._fw.policy.get_policy(policy).applied:
-                # first remove the old set of sources using the current zone
-                # settings.
-                if not enable and len(self.list_sources(zone)) == 1:
-                    self._fw.policy.unapply_policy_settings(policy, use_transaction=transaction)
-                else:
-                    self._fw.policy._ingress_egress_zones(False, policy, transaction)
-                    # after the transaction ends and therefore the sources
-                    # has been added to the zone's settings, update the
-                    # dependent policies
-                    transaction.add_post(lambda p: (p in self._fw.policy.get_active_policies_not_derived_from_zone()) and \
-                                                   self._fw.policy._ingress_egress_zones_transaction(True, p), policy)
-            elif enable:
-                transaction.add_post(lambda p: (p in self._fw.policy.get_active_policies_not_derived_from_zone()) and \
-                                               self._fw.policy.apply_policy_settings(p), policy)
+        self._interface_or_source_update_zone_termination(enable, zone, "", source, transaction)
+        self._interface_or_source_update_policies_derived_from_zone(enable, zone, "", source, transaction)
+        self._interface_or_source_update_policies(enable, zone, "", source, transaction)
 
     def add_service(self, zone, service, timeout=0, sender=None):
         zone = self._fw.check_zone(zone)
@@ -862,7 +909,7 @@ class FirewallZone(object):
         elif rule.element is None:
             return [self.policy_name_from_zones(zone, "HOST")]
         else:
-            raise FirewallError("Rich rule type (%s) not handled." % (type(rule.element)))
+            raise FirewallError(errors.INVALID_RULE, "Rich rule type (%s) not handled." % (type(rule.element)))
 
     def add_rule(self, zone, rule, timeout=0, sender=None):
         for p_name in self._rich_rule_to_policies(zone, rule):
@@ -1006,23 +1053,21 @@ class FirewallZone(object):
     def _forward(self, enable, zone, transaction):
         p_name = self.policy_name_from_zones(zone, "ANY")
 
-        for interface in self._zones[zone].settings["interfaces"]:
+        for interface in self._zones[zone].interfaces:
             for backend in self._fw.enabled_backends():
                 if not backend.policies_supported:
                     continue
                 rules = backend.build_zone_forward_rules(enable, zone, p_name, "filter", interface=interface)
                 transaction.add_rules(backend, rules)
 
-        for ipv,source in self._zones[zone].settings["sources"]:
+        for source in self._zones[zone].sources:
+            ipv = self.check_source(source)
             for backend in [self._fw.get_backend_by_ipv(ipv)] if ipv else self._fw.enabled_backends():
                 if not backend.policies_supported:
                     continue
                 rules = backend.build_zone_forward_rules(enable, zone, p_name, "filter", source=source)
                 transaction.add_rules(backend, rules)
 
-    def __forward_id(self):
-        return True
-
     def add_forward(self, zone, timeout=0, sender=None,
                     use_transaction=None):
         _zone = self._fw.check_zone(zone)
@@ -1030,8 +1075,7 @@ class FirewallZone(object):
         self._fw.check_panic()
         _obj = self._zones[_zone]
 
-        forward_id = self.__forward_id()
-        if forward_id in _obj.settings["forward"]:
+        if _obj.forward:
             raise FirewallError(errors.ALREADY_ENABLED,
                                 "forward already enabled in '%s'" % _zone)
 
@@ -1043,25 +1087,23 @@ class FirewallZone(object):
         if _obj.applied:
             self._forward(True, _zone, transaction)
 
-        self.__register_forward(_obj, forward_id, timeout, sender)
-        transaction.add_fail(self.__unregister_forward, _obj, forward_id)
+        self.__register_forward(_obj, timeout, sender)
+        transaction.add_fail(self.__unregister_forward, _obj)
 
         if use_transaction is None:
             transaction.execute(True)
 
         return _zone
 
-    def __register_forward(self, _obj, forward_id, timeout, sender):
-        _obj.settings["forward"][forward_id] = \
-            self.__gen_settings(timeout, sender)
+    def __register_forward(self, _obj, timeout, sender):
+        _obj.forward = True
 
     def remove_forward(self, zone, use_transaction=None):
         _zone = self._fw.check_zone(zone)
         self._fw.check_panic()
         _obj = self._zones[_zone]
 
-        forward_id = self.__forward_id()
-        if forward_id not in _obj.settings["forward"]:
+        if not _obj.forward:
             raise FirewallError(errors.NOT_ENABLED,
                                 "forward not enabled in '%s'" % _zone)
 
@@ -1073,16 +1115,15 @@ class FirewallZone(object):
         if _obj.applied:
             self._forward(False, _zone, transaction)
 
-        transaction.add_post(self.__unregister_forward, _obj, forward_id)
+        transaction.add_post(self.__unregister_forward, _obj)
 
         if use_transaction is None:
             transaction.execute(True)
 
         return _zone
 
-    def __unregister_forward(self, _obj, forward_id):
-        if forward_id in _obj.settings["forward"]:
-            del _obj.settings["forward"][forward_id]
+    def __unregister_forward(self, _obj):
+        _obj.forward = False
 
     def query_forward(self, zone):
-        return self.__forward_id() in self.get_settings(zone)["forward"]
+        return self.get_zone(zone).forward
