@@ -18,23 +18,20 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-# force use of pygobject3 in python-slip
-from gi.repository import GObject
-import sys
-sys.modules['gobject'] = GObject
 import os
 
 import dbus
 import dbus.service
-import slip.dbus
-import slip.dbus.service
 
 from firewall import config
 from firewall.core.base import DEFAULT_ZONE_TARGET
 from firewall.core.watcher import Watcher
 from firewall.core.logger import log
+from firewall.server.dbus import DbusServiceObject
 from firewall.server.decorators import handle_exceptions, \
-    dbus_handle_exceptions, dbus_service_method
+    dbus_handle_exceptions, dbus_service_method, \
+    dbus_service_method_deprecated, dbus_service_signal_deprecated, \
+    dbus_polkit_require_auth
 from firewall.server.config_icmptype import FirewallDConfigIcmpType
 from firewall.server.config_service import FirewallDConfigService
 from firewall.server.config_zone import FirewallDConfigZone
@@ -49,7 +46,8 @@ from firewall.core.io.direct import Direct
 from firewall.dbus_utils import dbus_to_python, \
     command_of_sender, context_of_sender, uid_of_sender, user_of_uid, \
     dbus_introspection_prepare_properties, \
-    dbus_introspection_add_properties
+    dbus_introspection_add_properties, \
+    dbus_introspection_add_deprecated
 from firewall import errors
 from firewall.errors import FirewallError
 
@@ -59,7 +57,7 @@ from firewall.errors import FirewallError
 #
 ############################################################################
 
-class FirewallDConfig(slip.dbus.service.Object):
+class FirewallDConfig(DbusServiceObject):
     """FirewallD main class"""
 
     persistent = True
@@ -97,20 +95,24 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.watcher.add_watch_file(config.FIREWALLD_DIRECT)
         self.watcher.add_watch_file(config.FIREWALLD_CONF)
 
-        dbus_introspection_prepare_properties(self,
-                                              config.dbus.DBUS_INTERFACE_CONFIG,
-                                              { "CleanupOnExit": "readwrite",
-                                                "IPv6_rpfilter": "readwrite",
-                                                "Lockdown": "readwrite",
-                                                "MinimalMark": "readwrite",
-                                                "IndividualCalls": "readwrite",
-                                                "LogDenied": "readwrite",
-                                                "AutomaticHelpers": "readwrite",
-                                                "FirewallBackend": "readwrite",
-                                                "FlushAllOnReload": "readwrite",
-                                                "RFC3964_IPv4": "readwrite",
-                                                "AllowZoneDrifting": "readwrite",
-                                              })
+        dbus_introspection_prepare_properties(self, config.dbus.DBUS_INTERFACE_CONFIG,
+            {
+                "CleanupOnExit": "readwrite",
+                "CleanupModulesOnExit": "readwrite",
+                "IPv6_rpfilter": "readwrite",
+                "Lockdown": "readwrite",
+                "MinimalMark": "readwrite",
+                "IndividualCalls": "readwrite",
+                "LogDenied": "readwrite",
+                "AutomaticHelpers": "readwrite",
+                "FirewallBackend": "readwrite",
+                "FlushAllOnReload": "readwrite",
+                "RFC3964_IPv4": "readwrite",
+                "AllowZoneDrifting": "readwrite",
+                "NftablesFlowtable": "readwrite",
+                "NftablesCounters": "readwrite",
+            }
+        )
 
     @handle_exceptions
     def _init_vars(self):
@@ -528,6 +530,13 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     @dbus_handle_exceptions
     def accessCheck(self, sender):
+        if self.config._fw._state == "FAILED":
+            raise FirewallError(errors.RUNNING_BUT_FAILED,
+                    "Changing permanent configuration is not allowed while "
+                    "firewalld is in FAILED state. The permanent "
+                    "configuration must be fixed and then firewalld "
+                    "restarted. Try `firewall-offline-cmd --check-config`.")
+
         if self.config.lockdown_enabled():
             if sender is None:
                 log.error("Lockdown not possible, sender not set.")
@@ -554,10 +563,10 @@ class FirewallDConfig(slip.dbus.service.Object):
     @dbus_handle_exceptions
     def _get_property(self, prop):
         if prop not in [ "DefaultZone", "MinimalMark", "CleanupOnExit",
-                         "Lockdown", "IPv6_rpfilter", "IndividualCalls",
-                         "LogDenied", "AutomaticHelpers", "FirewallBackend",
-                         "FlushAllOnReload", "RFC3964_IPv4",
-                         "AllowZoneDrifting" ]:
+                         "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
+                         "IndividualCalls", "LogDenied", "AutomaticHelpers",
+                         "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
+                         "AllowZoneDrifting", "NftablesFlowtable", "NftablesCounters"]:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.InvalidArgs: "
                 "Property '%s' does not exist" % prop)
@@ -578,6 +587,10 @@ class FirewallDConfig(slip.dbus.service.Object):
             if value is None:
                 value = "yes" if config.FALLBACK_CLEANUP_ON_EXIT else "no"
             return dbus.String(value)
+        elif prop == "CleanupModulesOnExit":
+            if value is None:
+                value = "yes" if config.FALLBACK_CLEANUP_MODULES_ON_EXIT else "no"
+            return dbus.String(value)
         elif prop == "Lockdown":
             if value is None:
                 value = "yes" if config.FALLBACK_LOCKDOWN else "no"
@@ -614,6 +627,14 @@ class FirewallDConfig(slip.dbus.service.Object):
             if value is None:
                 value = "yes" if config.FALLBACK_ALLOW_ZONE_DRIFTING else "no"
             return dbus.String(value)
+        elif prop == "NftablesFlowtable":
+            if value is None:
+                value = config.FALLBACK_NFTABLES_FLOWTABLE
+            return dbus.String(value)
+        elif prop == "NftablesCounters":
+            if value is None:
+                value = "yes" if config.FALLBACK_NFTABLES_COUNTERS else "no"
+            return dbus.String(value)
 
     @dbus_handle_exceptions
     def _get_dbus_property(self, prop):
@@ -623,6 +644,8 @@ class FirewallDConfig(slip.dbus.service.Object):
             return dbus.Int32(self._get_property(prop))
         elif prop == "CleanupOnExit":
             return dbus.String(self._get_property(prop))
+        elif prop == "CleanupModulesOnExit":
+            return dbus.String(self._get_property(prop))
         elif prop == "Lockdown":
             return dbus.String(self._get_property(prop))
         elif prop == "IPv6_rpfilter":
@@ -641,6 +664,10 @@ class FirewallDConfig(slip.dbus.service.Object):
             return dbus.String(self._get_property(prop))
         elif prop == "AllowZoneDrifting":
             return dbus.String(self._get_property(prop))
+        elif prop == "NftablesFlowtable":
+            return dbus.String(self._get_property(prop))
+        elif prop == "NftablesCounters":
+            return dbus.String(self._get_property(prop))
         else:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.InvalidArgs: "
@@ -667,8 +694,6 @@ class FirewallDConfig(slip.dbus.service.Object):
                 "org.freedesktop.DBus.Error.UnknownInterface: "
                 "Interface '%s' does not exist" % interface_name)
 
-        return self._get_dbus_property(property_name)
-
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='s',
                          out_signature='a{sv}')
     @dbus_handle_exceptions
@@ -679,10 +704,10 @@ class FirewallDConfig(slip.dbus.service.Object):
         ret = { }
         if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
             for x in [ "DefaultZone", "MinimalMark", "CleanupOnExit",
-                       "Lockdown", "IPv6_rpfilter", "IndividualCalls",
-                       "LogDenied", "AutomaticHelpers", "FirewallBackend",
-                       "FlushAllOnReload", "RFC3964_IPv4",
-                       "AllowZoneDrifting" ]:
+                       "CleanupModulesOnExit", "Lockdown", "IPv6_rpfilter",
+                       "IndividualCalls", "LogDenied", "AutomaticHelpers",
+                       "FirewallBackend", "FlushAllOnReload", "RFC3964_IPv4",
+                       "AllowZoneDrifting", "NftablesFlowtable",  "NftablesCounters"]:
                 ret[x] = self._get_property(x)
         elif interface_name in [ config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                                  config.dbus.DBUS_INTERFACE_CONFIG_POLICIES ]:
@@ -694,7 +719,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         return dbus.Dictionary(ret, signature="sv")
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_CONFIG)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_CONFIG)
     @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature='ssv')
     @dbus_handle_exceptions
     def Set(self, interface_name, property_name, new_value, sender=None):
@@ -706,49 +731,42 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.accessCheck(sender)
 
         if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
-            if property_name in [ "CleanupOnExit", "Lockdown",
+            if property_name in [ "CleanupOnExit",
+                                  "CleanupModulesOnExit", "Lockdown",
                                   "IPv6_rpfilter", "IndividualCalls",
                                   "LogDenied",
                                   "FirewallBackend", "FlushAllOnReload",
-                                  "RFC3964_IPv4", "AllowZoneDrifting" ]:
-                if property_name in [ "CleanupOnExit", "Lockdown",
-                                      "IPv6_rpfilter", "IndividualCalls" ]:
+                                  "RFC3964_IPv4", "NftablesFlowtable",
+                                  "NftablesCounters"]:
+                if property_name in [ "CleanupOnExit", "CleanupModulesOnExit",
+                                      "Lockdown", "IPv6_rpfilter",
+                                      "IndividualCalls", "FlushAllOnReload",
+                                      "RFC3964_IPv4", "NftablesCounters"]:
                     if new_value.lower() not in [ "yes", "no",
                                                   "true", "false" ]:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "LogDenied":
+                elif property_name == "LogDenied":
                     if new_value not in config.LOG_DENIED_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "FirewallBackend":
+                elif property_name == "FirewallBackend":
                     if new_value not in config.FIREWALL_BACKEND_VALUES:
                         raise FirewallError(errors.INVALID_VALUE,
                                             "'%s' for %s" % \
                                             (new_value, property_name))
-                if property_name == "FlushAllOnReload":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "RFC3964_IPv4":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
-                if property_name == "AllowZoneDrifting":
-                    if new_value.lower() not in ["yes", "true", "no", "false"]:
-                        raise FirewallError(errors.INVALID_VALUE,
-                                            "'%s' for %s" % \
-                                            (new_value, property_name))
+                else:
+                    raise dbus.exceptions.DBusException(
+                        "org.freedesktop.DBus.Error.InvalidArgs: "
+                        "Property '%s' does not exist" % property_name)
 
                 self.config.get_firewalld_conf().set(property_name, new_value)
                 self.config.get_firewalld_conf().write()
                 self.PropertiesChanged(interface_name,
                                        { property_name: new_value }, [ ])
-            elif property_name in ["MinimalMark", "AutomaticHelpers"]:
+            elif property_name in ["MinimalMark", "AutomaticHelpers", "AllowZoneDrifting"]:
                 # deprecated fields. Ignore setting them.
                 pass
             else:
@@ -774,7 +792,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.PropertiesChanged('%s', '%s', '%s')",
                    interface_name, changed_properties, invalidated_properties)
 
-    @slip.dbus.polkit.require_auth(config.dbus.PK_ACTION_INFO)
+    @dbus_polkit_require_auth(config.dbus.PK_ACTION_INFO)
     @dbus_service_method(dbus.INTROSPECTABLE_IFACE, out_signature='s')
     @dbus_handle_exceptions
     def Introspect(self, sender=None): # pylint: disable=W0613
@@ -782,8 +800,16 @@ class FirewallDConfig(slip.dbus.service.Object):
 
         data = super(FirewallDConfig, self).Introspect(self.path,
                                                        self.busname.get_bus())
-        return dbus_introspection_add_properties(
-            self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+        data = dbus_introspection_add_properties(
+                    self, data, config.dbus.DBUS_INTERFACE_CONFIG)
+
+        for interface in [config.dbus.DBUS_INTERFACE_CONFIG_DIRECT]:
+            data = dbus_introspection_add_deprecated(
+                        self, data, interface,
+                        dbus_service_method_deprecated().deprecated,
+                        dbus_service_signal_deprecated().deprecated)
+
+        return data
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
@@ -802,7 +828,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     def setLockdownWhitelist(self, settings, sender=None): # pylint: disable=W0613
         log.debug1("config.policies.setLockdownWhitelist(...)")
         settings = dbus_to_python(settings)
-        self.config.get_policies().lockdown_whitelist.import_config(settings)
+        self.config.get_policies().lockdown_whitelist.import_config(settings, {})
         self.config.get_policies().lockdown_whitelist.write()
         self.LockdownWhitelistUpdated()
 
@@ -1397,6 +1423,7 @@ class FirewallDConfig(slip.dbus.service.Object):
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
     # DIRECT
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1405,6 +1432,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         log.debug1("config.direct.getSettings()")
         return self.config.get_direct().export_config()
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature=Direct.DBUS_SIGNATURE)
     @dbus_handle_exceptions
@@ -1412,10 +1440,11 @@ class FirewallDConfig(slip.dbus.service.Object):
         # returns list ipv, table, list of chains
         log.debug1("config.direct.update()")
         settings = dbus_to_python(settings)
-        self.config.get_direct().import_config(settings)
+        self.config.get_direct().import_config(settings, {})
         self.config.get_direct().write()
         self.Updated()
 
+    @dbus_service_signal_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_handle_exceptions
     def Updated(self):
@@ -1423,6 +1452,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # chain
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1442,6 +1472,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].append(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1461,6 +1492,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[0].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='b')
     @dbus_handle_exceptions
@@ -1473,6 +1505,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = tuple((ipv, table, chain))
         return idx in self.getSettings()[0]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='ss', out_signature='as')
     @dbus_handle_exceptions
@@ -1486,6 +1519,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[2])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sss)')
     @dbus_handle_exceptions
@@ -1495,6 +1529,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # rule
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1516,6 +1551,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].append(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias')
     @dbus_handle_exceptions
@@ -1537,6 +1573,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[1].remove(idx)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sssias', out_signature='b')
     @dbus_handle_exceptions
@@ -1551,6 +1588,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, table, chain, priority, args)
         return idx in self.getSettings()[1]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss')
     @dbus_handle_exceptions
@@ -1567,6 +1605,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 settings[1].remove(rule)
         self.update(tuple(settings))
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sss', out_signature='a(ias)')
     @dbus_handle_exceptions
@@ -1582,6 +1621,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append((idx[3], idx[4]))
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='', out_signature='a(sssias)')
     @dbus_handle_exceptions
@@ -1591,6 +1631,7 @@ class FirewallDConfig(slip.dbus.service.Object):
 
     # passthrough
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1609,6 +1650,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         self.update(settings)
 
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas')
     @dbus_handle_exceptions
@@ -1626,6 +1668,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         settings[2].remove(idx)
         self.update(settings)
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='sas', out_signature='b')
     @dbus_handle_exceptions
@@ -1637,6 +1680,7 @@ class FirewallDConfig(slip.dbus.service.Object):
         idx = (ipv, args)
         return idx in self.getSettings()[2]
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          in_signature='s', out_signature='aas')
     @dbus_handle_exceptions
@@ -1649,6 +1693,7 @@ class FirewallDConfig(slip.dbus.service.Object):
                 ret.append(idx[1])
         return ret
 
+    @dbus_service_method_deprecated(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT)
     @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
                          out_signature='a(sas)')
     @dbus_handle_exceptions
